@@ -1,7 +1,7 @@
 /*
  * Security checks of environment
  * Added from shadow-utils package
- * by Arkadiusz Mikiewicz <misiek@pld.ORG.PL>
+ * by Arkadiusz Miśkiewicz <misiek@pld.ORG.PL>
  *
  */
 
@@ -26,7 +26,6 @@ extern char **environ;
 #endif
 
 static char * const forbid[] = {
-        "_RLD_=",
         "BASH_ENV=",    /* GNU creeping featurism strikes again... */
         "ENV=",
         "HOME=",
@@ -57,13 +56,15 @@ sanitize_env(void)
         char **envp = environ;
         char * const *bad;
         char **cur;
-        char **move;
+        int last = 0;
+
+        for (cur = envp; *cur; cur++)
+                last++;
 
         for (cur = envp; *cur; cur++) {
                 for (bad = forbid; *bad; bad++) {
                         if (strncmp(*cur, *bad, strlen(*bad)) == 0) {
-                                for (move = cur; *move; move++)
-                                        *move = *(move + 1);
+                                last = remote_entry(envp, cur - envp, last);
                                 cur--;
                                 break;
                         }
@@ -76,8 +77,7 @@ sanitize_env(void)
                                 continue;
                         if (!strchr(*cur, '/'))
                                 continue;  /* OK */
-                        for (move = cur; *move; move++)
-                                *move = *(move + 1);
+                        last = remote_entry(envp, cur - envp, last);
                         cur--;
                         break;
                 }
@@ -100,10 +100,42 @@ char *safe_getenv(const char *arg)
 		return NULL;
 #endif
 #endif
-
-#ifdef HAVE___SECURE_GETENV
+#ifdef HAVE_SECURE_GETENV
+return secure_getenv(arg);
+#elif HAVE___SECURE_GETENV
 	return __secure_getenv(arg);
 #else
 	return getenv(arg);
 #endif
 }
+
+#ifdef TEST_PROGRAM
+int main(void)
+{
+	char *const *bad;
+	char copy[32];
+	char *p;
+	int retval = EXIT_SUCCESS;
+
+	for (bad = forbid; *bad; bad++) {
+		strcpy(copy, *bad);
+		p = strchr(copy, '=');
+		if (p)
+			*p = '\0';
+		setenv(copy, copy, 1);
+	}
+	sanitize_env();
+	for (bad = forbid; *bad; bad++) {
+		strcpy(copy, *bad);
+		p = strchr(copy, '=');
+		if (p)
+			*p = '\0';
+		p = getenv(copy);
+		if (p) {
+			warnx("%s was not removed", copy);
+			retval = EXIT_FAILURE;
+		}
+	}
+	return retval;
+}
+#endif
