@@ -38,6 +38,7 @@
 #include "islocal.h"
 #include "nls.h"
 #include "pathnames.h"
+#include "pwdutils.h"
 #include "setpwnam.h"
 #include "strutils.h"
 #include "xalloc.h"
@@ -67,9 +68,9 @@ struct sinfo {
 	char *shell;
 };
 
-
-static void __attribute__((__noreturn__)) usage (FILE *fp)
+static void __attribute__((__noreturn__)) usage(void)
 {
+	FILE *fp = stdout;
 	fputs(USAGE_HEADER, fp);
 	fprintf(fp, _(" %s [options] [<username>]\n"), program_invocation_short_name);
 
@@ -80,52 +81,73 @@ static void __attribute__((__noreturn__)) usage (FILE *fp)
 	fputs(_(" -s, --shell <shell>  specify login shell\n"), fp);
 	fputs(_(" -l, --list-shells    print list of shells and exit\n"), fp);
 	fputs(USAGE_SEPARATOR, fp);
-	fputs(_(" -u, --help     display this help and exit\n"), fp);
-	fputs(_(" -v, --version  output version information and exit\n"), fp);
-	fprintf(fp, USAGE_MAN_TAIL("chsh(1)"));
-	exit(fp == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
+	printf( " -u, --help           %s\n", USAGE_OPTSTR_HELP);
+	printf( " -v, --version        %s\n", USAGE_OPTSTR_VERSION);
+	printf(USAGE_MAN_TAIL("chsh(1)"));
+	exit(EXIT_SUCCESS);
 }
 
 /*
- *  get_shell_list () -- if the given shell appears in /etc/shells,
+ *  is_known_shell() -- if the given shell appears in /etc/shells,
  *	return true.  if not, return false.
- *	if the given shell is NULL, /etc/shells is outputted to stdout.
  */
-static int get_shell_list(const char *shell_name)
+static int is_known_shell(const char *shell_name)
 {
-	FILE *fp;
-	int found = 0;
-	char *buf = NULL;
-	size_t sz = 0;
-	ssize_t len;
+	char *s, ret = 0;
 
-	fp = fopen(_PATH_SHELLS, "r");
-	if (!fp) {
-		if (!shell_name)
-			warnx(_("No known shells."));
+	if (!shell_name)
 		return 0;
+
+	setusershell();
+	while ((s = getusershell())) {
+		if (strcmp(shell_name, s) == 0) {
+			ret = 1;
+			break;
+		}
 	}
-	while ((len = getline(&buf, &sz, fp)) != -1) {
-		/* ignore comments and blank lines */
-		if (*buf == '#' || len < 2)
-			continue;
-		/* strip the ending newline */
-		if (buf[len - 1] == '\n')
-			buf[len - 1] = 0;
-		/* check or output the shell */
-		if (shell_name) {
-			if (!strcmp(shell_name, buf)) {
-				found = 1;
-				break;
-			}
-		} else
-			printf("%s\n", buf);
-	}
-	fclose(fp);
-	free(buf);
-	return found;
+	endusershell();
+	return ret;
 }
 
+/*
+ *  print_shells () -- /etc/shells is outputted to stdout.
+ */
+static void print_shells(void)
+{
+	char *s;
+
+	while ((s = getusershell()))
+		printf("%s\n", s);
+	endusershell();
+}
+
+#ifdef HAVE_LIBREADLINE
+static char *shell_name_generator(const char *text, int state)
+{
+	static size_t len;
+	char *s;
+
+	if (!state) {
+		setusershell();
+		len = strlen(text);
+	}
+
+	while ((s = getusershell())) {
+		if (strncmp(s, text, len) == 0)
+			return xstrdup(s);
+	}
+	return NULL;
+}
+
+static char **shell_name_completion(const char *text,
+				    int start __attribute__((__unused__)),
+				    int end __attribute__((__unused__)))
+{
+	rl_attempted_completion_over = 1;
+	return rl_completion_matches(text, shell_name_generator);
+}
+#endif
+
 /*
  *  parse_argv () --
  *	parse the command line arguments, and fill in "pinfo" with any
@@ -145,17 +167,14 @@ static void parse_argv(int argc, char **argv, struct sinfo *pinfo)
 	while ((c = getopt_long(argc, argv, "s:lhuv", long_options, NULL)) != -1) {
 		switch (c) {
 		case 'v':
-			printf(UTIL_LINUX_VERSION);
-			exit(EXIT_SUCCESS);
+			print_version(EXIT_SUCCESS);
 		case 'u': /* deprecated */
 		case 'h':
-			usage(stdout);
+			usage();
 		case 'l':
-			get_shell_list(NULL);
+			print_shells();
 			exit(EXIT_SUCCESS);
 		case 's':
-			if (!optarg)
-				usage(stderr);
 			pinfo->shell = optarg;
 			break;
 		default:
@@ -164,8 +183,9 @@ static void parse_argv(int argc, char **argv, struct sinfo *pinfo)
 	}
 	/* done parsing arguments.  check for a username. */
 	if (optind < argc) {
-		if (optind + 1 < argc)
-			usage(stderr);
+		if (optind + 1 < argc) {
+			errx(EXIT_FAILURE, _("cannot handle multiple usernames"));
+		}
 		pinfo->username = argv[optind];
 	}
 }
@@ -178,16 +198,18 @@ static char *ask_new_shell(char *question, char *oldshell)
 {
 	int len;
 	char *ans = NULL;
-#ifndef HAVE_LIBREADLINE
+#ifdef HAVE_LIBREADLINE
+	rl_attempted_completion_function = shell_name_completion;
+#else
 	size_t dummy = 0;
 #endif
-
 	if (!oldshell)
 		oldshell = "";
-	printf("%s [%s]: ", question, oldshell);
+	printf("%s [%s]:", question, oldshell);
 #ifdef HAVE_LIBREADLINE
-	if ((ans = readline(NULL)) == NULL)
+	if ((ans = readline(" ")) == NULL)
 #else
+	putchar(' ');
 	if (getline(&ans, &dummy, stdin) < 0)
 #endif
 		return NULL;
@@ -213,7 +235,7 @@ static void check_shell(const char *shell)
 		errx(EXIT_FAILURE, _("\"%s\" is not executable"), shell);
 	if (illegal_passwd_chars(shell))
 		errx(EXIT_FAILURE, _("%s: has illegal characters"), shell);
-	if (!get_shell_list(shell)) {
+	if (!is_known_shell(shell)) {
 #ifdef ONLY_LISTED_SHELLS
 		if (!getuid())
 			warnx(_("Warning: \"%s\" is not listed in %s."), shell,
@@ -233,7 +255,7 @@ static void check_shell(const char *shell)
 
 int main(int argc, char **argv)
 {
-	char *oldshell;
+	char *oldshell, *pwbuf;
 	int nullshell = 0;
 	const uid_t uid = getuid();
 	struct sinfo info = { NULL };
@@ -243,16 +265,16 @@ int main(int argc, char **argv)
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
-	atexit(close_stdout);
+	close_stdout_atexit();
 
 	parse_argv(argc, argv, &info);
 	if (!info.username) {
-		pw = getpwuid(uid);
+		pw = xgetpwuid(uid, &pwbuf);
 		if (!pw)
 			errx(EXIT_FAILURE, _("you (user %d) don't exist."),
 			     uid);
 	} else {
-		pw = getpwnam(info.username);
+		pw = xgetpwnam(info.username, &pwbuf);
 		if (!pw)
 			errx(EXIT_FAILURE, _("user \"%s\" does not exist."),
 			     info.username);
@@ -304,7 +326,7 @@ int main(int argc, char **argv)
 		    _("running UID doesn't match UID of user we're "
 		      "altering, shell change denied"));
 	}
-	if (uid != 0 && !get_shell_list(oldshell)) {
+	if (uid != 0 && !is_known_shell(oldshell)) {
 		errno = EACCES;
 		err(EXIT_FAILURE, _("your shell is not in %s, "
 				    "shell change denied"), _PATH_SHELLS);
